#include "hw.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>

#define _BV(bit) (1 << (bit)) // _BV macro is not defined for atmega8 so define it here

// BTNx is 1 if respective button is pressed
#define BTN0() (bit_is_clear(PIND, 0))
#define BTN1() (bit_is_clear(PIND, 1))
#define BTN2() (bit_is_clear(PIND, 2))
#define BTN3() (bit_is_clear(PIND, 3))

#define SEC_MIN 0
#define SEC_MAX 99 * 60 + 59

volatile unsigned char freq1 = 0; // second digit of frequency display
volatile unsigned char freq10 = 0; // first digit of frequency display
volatile unsigned char activeseg = 0; // indicates active display segment (used for multiplexing)
volatile unsigned int sec = 20 * 60; // default timer countdown: 20min and counting down
volatile unsigned int btn0cycles = 0; // these 4 are here to support one-time push and push-and-hold on the buttons
volatile unsigned int btn1cycles = 0;
volatile unsigned int btn2cycles = 0;
volatile unsigned int btn3cycles = 0;

#define _s_A 0
#define _s_B 1
#define _s_C 2
#define _s_D 3
#define _s_E 4
#define _s_F 5
#define _s_G 6
#define _s_dot 7

const unsigned char segs[] =
{
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F), // 0
	_BV(_s_B) | _BV(_s_C), // 1
	_BV(_s_A) | _BV(_s_B) | _BV(_s_D) | _BV(_s_E) | _BV(_s_G), // 2
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_G), // 3
	_BV(_s_B) | _BV(_s_C) | _BV(_s_F) | _BV(_s_G), // 4
	_BV(_s_A) | _BV(_s_C) | _BV(_s_D) | _BV(_s_F) | _BV(_s_G), // 5
	_BV(_s_A) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G), // 6
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C), // 7
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G), // 8
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_F) | _BV(_s_G), // 9
};

void addOrSubtractSeconds(int num)
{
	int result = sec + num;
	if (result >= SEC_MAX)
	{
		result = SEC_MAX;
	}
	else if (result <= SEC_MIN)
	{
		result = SEC_MIN;
	}
	
	sec = result;
}

int main(void)
{
	unsigned int i = 0;
	DDRB = 0xFF; // PD0-7 are outputs
	DDRC = 0b00111111; // PC0-PC5 are outputs, PC6-PC7 are inputs (unused)
	DDRD = 0b10000000; // PD7: output to drive relay, PD0-PD3: buttons input
	
	// show empty display at the beginning
	PORTB = 0xFF;
	PORTC = 0xFF;
	
	// PORTB: segments
	// PORTC: common anodes
	
	// TIMER0 configuration
	TCCR0 |= _BV(CS01) | _BV(CS02); // configure clock source for TIMER0 on pin T0
	
	// TIMER1 configuration
	TIMSK |= _BV(TOIE1); // enable overflow interrupt
	TCCR1B |= _BV(CS11); // configure prescaler to 8 and start timer

	// TIMER2 (display multiplexing) configuration
	TIMSK |= _BV(TOIE2); // enable overflow interrupt
	TCCR2 |= _BV(CS21); // configure prescaler to 8 and start timer

	sei(); // enable global interrupts
	
	// turn on relay initially
	PORTD |= _BV(7);
		
	while (1)
	{
		// maximum that this code detects, is 255 Hz, then timer overflows - this is not handled here as I won't need it (maximum real world frequency is 99 Hz)
		TCNT0 = 0;
		_delay_ms(1000);
		i = TCNT0;
		freq1 = i % 10;
		freq10 = i / 10;
		
		// timer handling
		if (sec != 0)
		{
			addOrSubtractSeconds(-1);

			// turn on relay (useful if time was raised by buttons)
			PORTD |= _BV(7);
		}
		else
		{
			// turn off relay
			PORTD &= ~_BV(7);
		}
	}
}

// button handling
ISR(TIMER1_OVF_vect)
{
	if (BTN0())
	{
		if (btn0cycles == 0 || btn0cycles > 10)
		{
			addOrSubtractSeconds(60);
		}
		
		btn0cycles++;
	}
	else if (BTN1())
	{
		if (btn1cycles == 0 || btn1cycles > 10)
		{
			addOrSubtractSeconds(-60);
		}
		
		btn1cycles++;
	}
	else if (BTN2())
	{
		if (btn2cycles == 0 || btn2cycles > 10)
		{
			addOrSubtractSeconds(60 * 5);
		}
	
		btn2cycles++;
	}
	else if (BTN3())
	{
		if (btn3cycles == 0 || btn3cycles > 10)
		{
			addOrSubtractSeconds(-60 * 5);
		}
	
		btn3cycles++;
	}
	
	if (!BTN0())
	{
		btn0cycles = 0;
	}
	
	if (!BTN1())
	{
		btn1cycles = 0;
	}
	
	if (!BTN2())
	{
		btn2cycles = 0;
	}
	
	if (!BTN3())
	{
		btn3cycles = 0;
	}
}

// display multiplexing
ISR(TIMER2_OVF_vect)
{
	if (activeseg == 5)
	{
		activeseg = 0;
	} else {
		activeseg++;
	}

	switch (activeseg)
	{
		case 0:
			PORTB = ~segs[freq10];
			PORTC = 1;
		break;
		case 1:
			PORTB = ~segs[freq1];
			PORTC = 2;
		break;
		case 2:
			// sec1
			PORTB = ~segs[sec % 10];
			PORTC = 4;
		break;
		case 3:
			// sec10
			PORTB = ~segs[(sec % 60) / 10];
			PORTC = 8;

			// blinking double dot
			if ((sec & 1) == 0 || sec == 0)
			{
				PORTB &= ~_BV(_s_dot);
		    } else {
				PORTB |= _BV(_s_dot);
			}
		break;
		case 4:
			// min1
			PORTB = ~segs[(sec / 60) % 10];
			PORTC = 16;

			// blinking double dot
			if ((sec & 1) == 0 || sec == 0)
			{
				PORTB &= ~_BV(_s_dot);
		    } else {
				PORTB |= _BV(_s_dot);
			}
		break;
		case 5:
			// min10
			PORTB = ~segs[sec / 600];
			PORTC = 32;
		break;
	}
}
